using System;
using System.Collections;
using Ringlet.Misc.VCS;

namespace Ringlet.Collections
{
	/// <summary>
	/// A deque implemented as a fixed-size array.
	/// </summary>
	[VCSAttribute("$Ringlet: cs/misc/RingletCollections/Ringlet.Collections/RotDeque.cs,v 1.6 2004/04/09 13:22:33 roam Exp $")]
	public class RotDeque : ICollection, IEnumerable, IList, ICloneable
	{
		/// <summary>
		/// The default size for a newly-created deque.
		/// </summary>
		public const int RotDequeDefaultSize = 16;

		/// <summary>
		/// The deque size.
		/// </summary>
		protected int size;
		/// <summary>
		/// The items in the deque.
		/// </summary>
		protected object[] items;
		/// <summary>
		/// The element at the head of the deque.
		/// </summary>
		protected int head;
		/// <summary>
		/// The element just past the tail of the deque.
		/// </summary>
		protected int tail;

		/// <summary>
		/// Initialize a RotDeque object with a default size of 16.
		/// </summary>
		public RotDeque()
			: this(0)
		{
		}

		/// <summary>
		/// Initialize a RotDeque object with the specified size.
		/// </summary>
		/// <param name="sz">the deque size</param>
		/// <exception cref="System.ArgumentException">
		/// A negative value was specified for the <paramref
		/// name="size">size</paramref> parameter.
		/// </exception>
		public RotDeque(int sz)
		{
			if (sz < 0)
				throw (new ArgumentException(
				    "Negative size specified for a RotDeque",
				    "sz"));
			if (sz == 0)
				sz = RotDequeDefaultSize;
			size = sz + 1;
			items = new object[size];
			head = tail = 0;
		}

		/// <summary>
		/// Query whether the deque is empty.
		/// </summary>
		public virtual bool Empty
		{
			get
			{
				return (head == tail);
			}
		}

		/// <summary>
		/// Query whether the deque is full.
		/// </summary>
		public virtual bool Full
		{
			get
			{
				return (Count == size - 1);
			}
		}

		/// <summary>
		/// Get the number of elements currently on the deque.
		/// </summary>
		public virtual int Count
		{
			get
			{
				return ((tail + size - head) % size);
			}
		}

		/// <summary>
		/// Get the deque's capacity.
		/// </summary>
		public virtual int Capacity
		{
			get
			{
				return (size - 1);
			}
		}

		/// <summary>
		/// Get the actual position of an object in the items[] array.
		/// </summary>
		/// <param name="idx">the index relative to the deque's head</param>
		/// <returns>The index in the items[] array.</returns>
		protected int RealIndex(int idx)
		{
			return ((head + idx) % size);
		}

		/// <summary>
		/// Retrieve an element from the deque.
		/// </summary>
		/// <exception cref="System.ArgumentOutOfRangeException">
		/// The <paramref name="idx">idx</paramref> parameter was
		/// either negative or larger than the current element count.
		/// </exception>
		public virtual object this[int idx]
		{
			get
			{
				if ((idx < 0) || (idx >= Count))
					throw (new ArgumentOutOfRangeException(
					    "idx", idx,
					    "Invalid RotDeque index specified"));
				return (items[RealIndex(idx)]);
			}

			set
			{
				if ((idx < 0) || (idx >= Count))
					throw (new ArgumentOutOfRangeException(
					    "idx", idx,
					    "Invalid RotDeque index specified"));
				items[RealIndex(idx)] = value;
			}
		}

		/// <summary>
		/// Push an element on the head of the deque.
		/// </summary>
		/// <param name="o">The object to push.</param>
		/// <exception cref="System.InvalidOperationException">
		/// An attempt was made to push an element onto a full deque.
		/// </exception>
		public virtual void PushFront(object o)
		{
			if (Full)
				throw (new InvalidOperationException("The deque is full"));
			if (head == 0)
				head = size - 1;
			else
				head--;
			items[head] = o;
		}

		/// <summary>
		/// Push an element at the end of the deque.
		/// </summary>
		/// <param name="o">The object to push.</param>
		/// <exception cref="System.InvalidOperationException">
		/// An attempt was made to push an element onto a full deque.
		/// </exception>
		public virtual void PushBack(object o)
		{
			if (Full)
				throw (new InvalidOperationException("The deque is full"));
			items[tail] = o;
			if (tail == size - 1)
				tail = 0;
			else
				tail++;
		}

		/// <summary>
		/// Pop an element from the head of the deque.
		/// </summary>
		/// <returns>The element removed from the head of the deque.</returns>
		/// <exception cref="System.InvalidOperationException">
		/// An attempt was made to pop an element from an empty deque.
		/// </exception>
		public virtual object PopFront()
		{
			object o;

			if (Empty)
				throw (new InvalidOperationException("The deque is empty"));
			o = items[head];
			items[head] = null;
			if (head == size - 1)
				head = 0;
			else
				head++;
			return (o);
		}

		/// <summary>
		/// Pop an element from the back of the deque.
		/// </summary>
		/// <returns>The element removed from the back of the deque.</returns>
		/// <exception cref="System.InvalidOperationException">
		/// An attempt was made to pop an element from an empty deque.
		/// </exception>
		public virtual object PopBack()
		{
			object o;

			if (Empty)
				throw (new InvalidOperationException("The deque is empty"));
			if (tail == 0)
				tail = size - 1;
			else
				tail--;
			o = items[tail];
			items[tail] = null;
			return (o);
		}

		/// <summary>
		/// Non-destructive retrieval of the element at the head
		/// of the deque.
		/// </summary>
		/// <returns>The element at the head of the deque.</returns>
		/// <exception cref="System.InvalidOperationException">
		/// An attempt was made to fetch an element from an empty deque.
		/// </exception>
		public virtual object AtFront
		{
			get
			{
				if (Empty)
					throw (new InvalidOperationException(
					    "The deque is empty"));
				return (items[head]);
			}
		}

		/// <summary>
		/// Non-destructive retrieval of the element at the back
		/// of the deque.
		/// </summary>
		/// <returns>The element at the back of the deque.</returns>
		/// <exception cref="System.InvalidOperationException">
		/// An attempt was made to fetch an element from an empty deque.
		/// </exception>
		public virtual object AtBack
		{
			get
			{
				if (Empty)
					throw (new InvalidOperationException(
					    "The deque is empty"));
				if (tail == 0)
					return (items[size - 1]);
				else
					return (items[tail - 1]);
			}
		}

		/// <summary>
		/// Gets a value indicating whether access to the RotDeque
		/// is synchronized (thread-safe).
		/// </summary>
		public virtual bool IsSynchronized
		{
			get
			{
				return (false);
			}
		}

		/// <summary>
		/// Gets an object that may be used to synchronize access
		/// to the RotDeque.
		/// </summary>
		public virtual object SyncRoot
		{
			get
			{
				return (this);
			}
		}

		/// <summary>
		/// Gets an enumerator on the deque's elements.
		/// </summary>
		/// <returns>An enumerator.</returns>
		public virtual IEnumerator GetEnumerator()
		{
			return (new RotDequeEnumerator(this));
		}

		/// <summary>
		/// Copies the deque's contents to an array.
		/// </summary>
		/// <param name="array">The array to copy to.</param>
		/// <param name="idx">The array position to start copying
		/// at.</param>
		/// <exception cref="System.ArgumentException">
		/// A multidimensional <paramref name="array">array</paramref>
		/// parameter was specified, or there was not enough space in
		/// the array to accomodate all of the deque's elements.
		/// </exception>
		/// <exception cref="System.ArgumentNullException">
		/// An undefined value was specified for the <paramref
		/// name="array">array</paramref> parameter.
		/// </exception>
		/// <exception cref="System.ArgumentOutOfRangeException">
		/// A negative value was specified for the <paramref
		/// name="idx">idx</paramref> parameter.
		/// </exception>
		public virtual void CopyTo(Array array, int idx)
		{

			if (array == null)
				throw (new ArgumentNullException("array"));
			if (idx < 0)
				throw (new ArgumentOutOfRangeException("idx"));
			if (array.Rank != 1)
				throw (new ArgumentException(
				    "Cannot copy a RotDeque to a multidimensional array",
				    "array"));
			if (idx + Count > array.Length)
				throw (new ArgumentException(
				    "Not enough space left in the array to accomodate the RotDeque's elements",
				    "idx"));

			// OK, roll with it...
			if (head == tail)
				return;
			if (head < tail)
			{
				Array.Copy(items, head, array, idx, tail - head);
			}
			else
			{
				Array.Copy(items, head, array, idx, size - head);
				Array.Copy(items, 0, array, idx + size - head, tail);
			}
		}

		/// <summary>
		/// Push the deque's elements to the left (so that
		/// <see href="head" /> = 0).
		/// </summary>
		public virtual void Normalize()
		{
			int ofs;
			object[] n;

			if (head == 0)
				return;
			if (Empty)
			{
				head = tail = 0;
				return;
			}
			if (tail >= head)
			{
				Array.Copy(items, head, items, 0, tail - head);
				for (ofs = tail - head; ofs < tail; ofs++)
					items[ofs] = null;
			}
			else
			{
				ofs = size - head;
				n = new object[size];
				Array.Copy(items, head, n, 0, ofs);
				Array.Copy(items, 0, n, ofs, tail);
				items = n;
			}
			tail = Count;
			head = 0;
		}

		/// <summary>
		/// Append the elements of a deque to the end of another deque.
		/// </summary>
		/// <param name="d">The deque to append to.</param>
		/// <exception cref="System.ArgumentNullException">
		/// The <paramref name="d">d</paramref> parameter had an
		/// undefined value.
		/// </exception>
		/// <exception cref="System.ArgumentException">
		/// There was not enough space in the destination deque
		/// <paramref name="d">d</paramref> to accomodate all of the
		/// source deque's elements.
		/// </exception>
		public virtual void CopyTo(RotDeque d)
		{
			
			if (d == null)
				throw (new ArgumentNullException("array"));
			if (d.Count + Count > d.Capacity)
				throw (new ArgumentException(
				    "Not enough space left in the deque to accomodate the RotDeque's elements",
				    "d"));
			if (Empty)
				return;
			Normalize();
			d.Normalize();
			Array.Copy(items, 0, d.items, d.tail, tail);
			d.tail += tail;
		}

		/// <summary>
		/// Insert an item into the deque at a specified position.
		/// </summary>
		/// <param name="pos">The position to insert at.</param>
		/// <param name="o">The element to insert.</param>
		/// <remarks>The position is zero-based, with 0 being
		/// equivalent to a PushFront(), and Count being equivalent to
		/// a PushBack().</remarks>
		/// <exception cref="System.InvalidOperationException">
		/// An attempt was made to insert an element into a full deque.
		/// </exception>
		/// <exception cref="System.ArgumentOutOfRangeException">
		/// The <paramref name="pos">pos</paramref> parameter specified
		/// a position beyond the deque boundaries.
		/// </exception>
		public virtual void Insert(int pos, object o)
		{
			int idx, i;

			if (Full)
				throw (new InvalidOperationException(
				    "Cannot insert into a full deque"));
			if (pos > Count)
				throw (new ArgumentOutOfRangeException(
				    "pos", pos,
				    "Invalid position to insert into the deque at"));
			if (pos == Count)
			{
				PushBack(o);
				return;
			}
			else if (pos == 0)
			{
				PushFront(o);
				return;
			}

			// OK, some real work to do now
			idx = RealIndex(pos);
			if (idx < tail)
			{
				for (i = tail; i > idx; i--)
					items[i] = items[i - 1];
			}
			else
			{
				for (i = tail; i > 0; i--)
					items[i] = items[i - 1];
				items[0] = items[size - 1];
				for (i = size - 1; i > idx; i--)
					items[i] = items[i - 1];
			}
			items[idx] = o;
			tail++;
		}

		/// <summary>
		/// Add an element to the end of the deque.
		/// </summary>
		/// <param name="o">The element to add.</param>
		/// <returns>The index of the element added.</returns>
		public virtual int Add(object o)
		{
			PushBack(o);
			return (Count - 1);
		}

		/// <summary>
		/// Remove all elements from the deque.
		/// </summary>
		public virtual void Clear()
		{
			int i;

			for (i = 0; i < size; i++)
				items[i] = null;
			head = tail = 0;
		}

		/// <summary>
		/// Determine whether an element is on the deque.
		/// </summary>
		/// <param name="o">The element to look for.</param>
		/// <returns>True if the element is on the deque.</returns>
		public virtual bool Contains(object o)
		{
			return (IndexOf(o) != -1);
		}

		/// <summary>
		/// Determine the index of an element on the deque.
		/// </summary>
		/// <param name="o">The element to look for.</param>
		/// <returns>The index of the element on the deque, or -1 if
		/// the element is not present.</returns>
		public virtual int IndexOf(object o)
		{
			int i;

			for (i = head; i != tail; i++)
			{
				if (items[i] == o)
					return ((i + size - head) % size);
				// Naughty naughty
				if (i == size - 1)
					i = -1;
			}
			return (-1);
		}

		/// <summary>
		/// Get a value indicating whether the deque has a fixed size.
		/// </summary>
		public virtual bool IsFixedSize
		{
			get
			{
				return (false);
			}
		}

		/// <summary>
		/// Get a value indicating whether the deque is read-only.
		/// </summary>
		public virtual bool IsReadOnly
		{
			get
			{
				return (false);
			}
		}

		/// <summary>
		/// Remove the first occurence of an element from the deque.
		/// </summary>
		/// <param name="o">The element to remove.</param>
		public virtual void Remove(object o)
		{
			int idx;

			idx = IndexOf(o);
			if (idx == -1)
				return;
			RemoveAt(idx);
		}

		/// <summary>
		/// Remove the element at a specified position on the deque.
		/// </summary>
		/// <param name="idx">The position of the element to remove.
		/// </param>
		/// <exception cref="System.ArgumentOutOfRangeException">
		/// A negative or too large index was specified by the
		/// <paramref name="idx">idx</paramref> parameter.
		/// </exception>
		public virtual void RemoveAt(int idx)
		{
			int i;

			if (idx < 0 || idx >= Count)
				throw (new ArgumentOutOfRangeException("idx", idx,
				    "Invalid index to remove from the deque"));
			idx = RealIndex(idx);
			if (idx == head)
			{
				PopFront();
				return;
			}
			else if (((idx + 1) % size) == tail)
			{
				PopBack();
				return;
			}

			if (idx < tail)
			{
				for (i = idx; i < tail; i++)
					items[i] = items[i + 1];
			}
			else
			{
				for (i = idx; i < size - 1; i++)
					items[i] = items[i + 1];
				items[i] = items[0];
				for (i = 0; i < tail; i++)
					items[i] = items[i + 1];
			}
			items[tail] = null;
			if (tail == 0)
				tail = size - 1;
			else
				tail--;
		}

		/// <summary>
		/// Perform a shallow copy of the deque.
		/// </summary>
		/// <returns>A shallow copy of the deque.</returns>
		public object Clone()
		{
			RotDeque d;

			d = new RotDeque(size - 1);
			items.CopyTo(d.items, 0);
			d.head = head;
			d.tail = tail;
			return (d);
		}

		/// <summary>
		/// Copy the deque's contents into an array of deques.
		/// </summary>
		/// <param name="deques">The destination array of deques.</param>
		/// <exception cref="System.ArgumentNullException">
		/// Either the destination array or some of the deques within
		/// it had an undefined value.
		/// </exception>
		/// <exception cref="System.ArgumentException">
		/// There was not enough total free space in the destination
		/// array of deques.
		/// </exception>
		/// <exception cref="System.ApplicationException">
		/// The free space in the deques was calculated to be enough
		/// at the start of the operation, but either the source deque
		/// or some of the destination deques were modified in-flight,
		/// and not all elements could be copied.
		/// </exception>
		public virtual void Split(RotDeque[] deques)
		{
			int i, j, ofs, count;

			if (deques == null)
				throw (new ArgumentNullException("deques",
				    "Cannot split into a null deques array"));

			count = deques.Length;
			Normalize();

			// Check for free space in the destination deques
			ofs = 0;
			for (i = 0; i < count; i++)
			{
				if (deques[i] == null)
					throw (new ArgumentNullException(
					    String.Format("deques[{0}]", i),
					    "Null destination deque for splitting"));
				deques[i].Normalize();
				ofs += deques[i].size - deques[i].tail - 1;
			}
			if (ofs < tail)
				throw (new ArgumentException(String.Format(
				    "Not enough space left in the destination deques: have {0}, need {1}",
				    ofs, Count), "deques"));

			// Roll with it :)
			for (i = 0, j = 0; i < count && j < tail; i++)
			{
				ofs = deques[i].size - deques[i].tail - 1;
				Array.Copy(items, j, deques[i].items,
				    deques[i].tail, ofs);
				deques[i].tail += ofs;
				j += ofs;
			}
			// I *think* we're done.
			if (j < tail)
				throw (new ApplicationException(
				    "Some of the deques were modified during the splitting, the operation could not complete."));
		}

		/// <summary>
		/// An enumerator for the elements of a RotDeque
		/// </summary>
		[VCSAttribute("$Ringlet: cs/misc/RingletCollections/Ringlet.Collections/RotDeque.cs,v 1.6 2004/04/09 13:22:33 roam Exp $")]
		protected sealed class RotDequeEnumerator : IEnumerator
		{
			/// <summary>
			/// The deque to enumerate.
			/// </summary>
			private RotDeque deque;
			/// <summary>
			/// The position to start enumerating from.
			/// </summary>
			private int pos;

			/// <summary>
			/// Initialize a RotDequeEnumerator from the start
			/// of a deque.
			/// </summary>
			/// <param name="d">The deque to enumerate.</param>
			public RotDequeEnumerator(RotDeque d)
			{
				deque = d;
				pos = -1;
			}

			/// <summary>
			/// Initialize a RotDequeEnumerator from a specified
			/// position of the deque.
			/// </summary>
			/// <param name="d"></param>
			/// <param name="p"></param>
			public RotDequeEnumerator(RotDeque d, int p)
			{
				deque = d;
				pos = p;
			}

			/// <summary>
			/// Retrieve the current object.
			/// </summary>
			object IEnumerator.Current
			{
				get
				{
					return (deque[pos]);
				}
			}

			/// <summary>
			/// Advance the enumerator if possible.
			/// </summary>
			/// <returns>True if there are more elements to
			/// be retrieved.</returns>
			bool IEnumerator.MoveNext()
			{
				if (pos >= deque.Count - 1)
					return (false);
				pos++;
				return (true);
			}

			/// <summary>
			/// Reset the enumerator to the head of the deque.
			/// </summary>
			void IEnumerator.Reset()
			{
				pos = -1;
			}
		}
	}
}
